package activitypub

import (
	"context"
	"database/sql"
	"encoding/json"
	"fmt"
	"io"
	"net/http"
	"regexp"
	"slices"
	"strconv"
	"strings"
	"time"

	webutils "git.mstar.dev/mstar/goutils/http"
	"git.mstar.dev/mstar/goutils/other"
	"git.mstar.dev/mstar/goutils/sliceutils"
	"github.com/mitchellh/mapstructure"
	"github.com/rs/zerolog/hlog"
	"github.com/rs/zerolog/log"
	"gorm.io/gorm"

	"git.mstar.dev/mstar/linstrom/activitypub"
	"git.mstar.dev/mstar/linstrom/activitypub/translators"
	"git.mstar.dev/mstar/linstrom/config"
	"git.mstar.dev/mstar/linstrom/shared"
	"git.mstar.dev/mstar/linstrom/storage-new"
	"git.mstar.dev/mstar/linstrom/storage-new/dbgen"
	"git.mstar.dev/mstar/linstrom/storage-new/models"
	webshared "git.mstar.dev/mstar/linstrom/web/shared"
)

var objectIdRegex = regexp.MustCompile(
	`https?://.+/api/activitypub/[a-z]+(?:/[a-z]+)?/([a-zA-Z0-9\-]+)`,
)

func userInbox(w http.ResponseWriter, r *http.Request) {
	log := hlog.FromRequest(r)
	userId := r.PathValue("id")
	body, err := io.ReadAll(r.Body)
	log.Trace().
		Err(err).
		Str("userId", userId).
		Bytes("body", body).
		Any("headers", r.Header).
		Msg("Inbox message")
	data := map[string]any{}
	err = json.Unmarshal(body, &data)
	if err != nil {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer("Body to inbox needs to be json"),
			nil,
		)
		return
	}
	if _, ok := data["@context"]; !ok {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer("Request data needs to contain context"),
			nil,
		)
		return
	}
	objectType, ok := data["type"].(string)
	if !ok {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer(`Request data needs to contain a field "type" with a string value`),
			nil,
		)
		return
	}
	_, ok = data["id"].(string)
	if !ok {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer(`Request data needs to contain a field "id" with a string value`),
			nil,
		)
		return
	}
	log.Debug().Str("object-type", objectType).Msg("Inbox message")
	log.Trace().Bytes("body", body).Msg("Inbox raw message")
	// TODO: Decide how to handle the handler failing for whatever reason
	// Add object to unhandled message table and try again later?
	// Discard it? And how would a handler return that it failed?
	ok = true
	switch objectType {
	case "Like":
		ok = handleLike(w, r, data)
	case "Undo":
		ok = handleUndo(w, r, data)
	case "Follow":
		ok = handleFollow(w, r, data)
	case "Accept":
		ok = handleAccept(w, r, data)
	case "Reject":
		ok = handleReject(w, r, data)
	case "Create":
		ok = handleCreate(w, r, data)
	default:
		log.Warn().
			Str("object-type", objectType).
			Msg("Unknown message type, storing for later processing")
		err = dbgen.UnhandledMessage.Create(&models.UnhandledMessage{
			ForUserId:   userId,
			GlobalInbox: false,
			RawData:     body,
		})
		if err != nil {
			log.Error().Err(err).Msg("Failed to store unhandled message for later")
		}
		_ = webutils.ProblemDetailsStatusOnly(w, 500)
	}
	if !ok {
		err = dbgen.UnhandledMessage.Create(&models.UnhandledMessage{
			RawData:     body,
			ForUserId:   userId,
			GlobalInbox: false,
		})
		if err != nil {
			log.Error().
				Err(err).
				Bytes("body", body).
				Str("user-id", userId).
				Msg("Failed to store failed inbound message for later processing")
		}
	}
}

func globalInbox(w http.ResponseWriter, r *http.Request) {
	log := hlog.FromRequest(r)
	body, err := io.ReadAll(r.Body)
	log.Trace().
		Err(err).
		Bytes("body", body).
		Any("headers", r.Header).
		Msg("Global inbox message")
	data := map[string]any{}
	err = json.Unmarshal(body, &data)
	if err != nil {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer("Body to inbox needs to be json"),
			nil,
		)
		return
	}
	if _, ok := data["@context"]; !ok {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer("Request data needs to contain context"),
			nil,
		)
		return
	}
	objectType, ok := data["type"].(string)
	if !ok {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer(`Request data needs to contain a field "type" with a string value`),
			nil,
		)
		return
	}
	_, ok = data["id"].(string)
	if !ok {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer(`Request data needs to contain a field "id" with a string value`),
			nil,
		)
		return
	}
	log.Debug().Str("object-type", objectType).Msg("Inbox message")
	log.Trace().Bytes("body", body).Msg("Inbox raw message")
	// TODO: Decide how to handle the handler failing for whatever reason
	// Add object to unhandled message table and try again later?
	// Discard it? And how would a handler return that it failed?
	ok = true
	switch objectType {
	case "Like":
		ok = handleLike(w, r, data)
	case "Undo":
		ok = handleUndo(w, r, data)
	case "Accept":
		ok = handleAccept(w, r, data)
	case "Reject":
		ok = handleReject(w, r, data)
	case "Create":
		ok = handleCreate(w, r, data)
	default:
		log.Warn().
			Str("object-type", objectType).
			Msg("Unknown message type, storing for later processing")
		err = dbgen.UnhandledMessage.Create(&models.UnhandledMessage{
			ForUserId:   "",
			GlobalInbox: true,
			RawData:     body,
		})
		if err != nil {
			log.Error().Err(err).Msg("Failed to store unhandled message for later")
		}
		_ = webutils.ProblemDetailsStatusOnly(w, 500)
	}
	if !ok {
		err = dbgen.UnhandledMessage.Create(&models.UnhandledMessage{
			RawData:     body,
			ForUserId:   "",
			GlobalInbox: true,
		})
		if err != nil {
			log.Error().
				Err(err).
				Bytes("body", body).
				Msg("Failed to store failed global inbound message for later processing")
		}
	}
}

func handleLike(w http.ResponseWriter, r *http.Request, object map[string]any) bool {
	log := hlog.FromRequest(r)
	activityId := object["id"].(string)
	likerUrl, ok := object["actor"].(string)
	if !ok {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer(`Request data needs to contain a field "actor" with a string value`),
			nil,
		)
		return true
	}
	// TODO: Account for case where object is embedded in like
	targetUrl, ok := object["object"].(string)
	if !ok {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer(`Request data needs to contain a field "object" with a string value`),
			nil,
		)
		return true
	}
	targetIdMatches := objectIdRegex.FindStringSubmatch(targetUrl)
	if len(targetIdMatches) != 2 {
		log.Error().
			Strs("match-results", targetIdMatches).
			Str("url", targetUrl).
			Msg("Url didn't match id extractor regex")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return true
	}
	targetId := targetIdMatches[1]
	// Assume likes can only happen on notes for now
	// Thus check if a note with that Id exists at all
	note, err := dbgen.Note.Where(dbgen.Note.ID.Eq(targetId)).First()
	switch err {
	case nil:
	case gorm.ErrRecordNotFound:
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer("There is no note with the target id"),
			nil,
		)
		return true
	default:
		log.Error().Err(err).Str("note-id", targetId).Msg("Failed to get note from db")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	// Import liker after verifying that target note is correct
	liker, err := activitypub.ImportRemoteAccountByAPUrl(likerUrl)
	if err != nil {
		log.Error().
			Err(err).
			Str("liker-url", likerUrl).
			Msg("Failed to import liking remote account")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	reaction := models.Reaction{
		Note:      *note,
		NoteId:    note.ID,
		Reactor:   *liker,
		ReactorId: liker.ID,
		Emote:     nil,
		EmoteId:   sql.NullString{Valid: false},
	}
	tx := dbgen.Q.Begin()

	err = tx.Reaction.Create(&reaction)
	if err != nil {
		_ = tx.Rollback()
		log.Error().Err(err).Any("raw-reaction", reaction).Msg("Failed to store reaction in db")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	// TODO: Create corresponding activity too
	activity := models.Activity{
		Id:         activityId,
		Type:       string(models.ActivityLike),
		ObjectId:   fmt.Sprint(reaction.ID),
		ObjectType: uint32(models.ActivitystreamsActivityTargetReaction),
	}
	err = tx.Activity.Create(&activity)
	if err != nil {
		_ = tx.Rollback()
		log.Error().Err(err).Any("raw-reaction", reaction).Msg("Failed to store reaction in db")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	err = tx.Commit()
	if err != nil {
		log.Error().Err(err).Msg("Failed to commit reaction transaction to db")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	return true
}

func handleFollow(w http.ResponseWriter, r *http.Request, object map[string]any) bool {
	log := hlog.FromRequest(r)
	log.Debug().Msg("Received follow request")
	objectId, ok := object["id"].(string)
	if !ok {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer(`Request data needs to contain a field "id" with a string value`),
			nil,
		)
		return true
	}
	actorApId, ok := object["actor"].(string)
	if !ok {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer(`Request data needs to contain a field "actor" with a string value`),
			nil,
		)
		return true
	}
	targetUrl, ok := object["object"].(string)
	if !ok {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer(`Request data needs to contain a field "object" of type string`),
			nil,
		)
		return true
	}
	followedMatch := objectIdRegex.FindStringSubmatch(targetUrl)
	if len(followedMatch) != 2 {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer(`Object must be a link to a Linstrom AP user`),
			nil,
		)
		return true
	}
	followedId := followedMatch[1]
	followed, err := dbgen.User.Where(dbgen.User.ID.Eq(followedId)).First()
	switch err {
	case gorm.ErrRecordNotFound:
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusNotFound)
		return true
	case nil:
	default:
		log.Error().Err(err).Str("target-id", followedId).Msg("Failed to get account from db")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	follower, err := activitypub.ImportRemoteAccountByAPUrl(actorApId)
	if err != nil {
		log.Error().Err(err).Msg("Failed to import following account")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	u2u := dbgen.UserToUserRelation
	followRelations, err := u2u.Where(
		u2u.UserId.Eq(follower.ID),
		u2u.TargetUserId.Eq(followed.ID),
		u2u.Or(
			u2u.Relation.Eq(string(models.RelationFollow)),
			u2u.Relation.Eq(string(models.RelationFollowRequest)),
		),
	).Count()
	if err != nil {
		log.Error().
			Err(err).
			Str("follower", follower.ID).
			Str("followed", followedId).
			Msg("Failed to count follow relations")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	if followRelations > 0 {
		return true
	}
	tx := dbgen.Q.Begin()
	req := models.UserToUserRelation{
		Relation:     string(models.RelationFollowRequest),
		TargetUserId: followed.ID,
		UserId:       follower.ID,
	}
	err = tx.UserToUserRelation.Create(&req)
	if err != nil {
		_ = tx.Rollback()
		log.Error().Err(err).Any("follow-request", req).Msg("Failed to store follow request")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	activity := models.Activity{
		Id:         objectId,
		Type:       string(models.ActivityFollow),
		ObjectId:   fmt.Sprint(req.ID),
		ObjectType: uint32(models.ActivitystreamsActivityTargetFollow),
	}
	err = tx.Activity.Create(&activity)
	if err != nil {
		_ = tx.Rollback()
		log.Error().Err(err).Any("activity", activity).Msg("Failed to store follow activity")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	err = tx.Commit()
	if err != nil {
		log.Error().Err(err).Msg("Failed to commit follow activity transaction")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	if !followed.RestrictedFollow {
		go func() {
			err := AcceptFollow(r.Context(), req.ID, followedId, followedId)
			if err != nil {
				log.Error().
					Err(err).
					Uint64("follow-request-id", req.ID).
					Msg("Failed to auto-accept follow request")
			}
		}()
	}
	return true
}

// WARN: Untested as can't send follow activities yet
func handleAccept(w http.ResponseWriter, r *http.Request, object map[string]any) bool {
	log := hlog.FromRequest(r)
	rawTarget, ok := object["object"]
	if !ok {
		log.Debug().Msg("Rejecting accept, missing object")
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer(`Request data needs to contain a field "object"`),
			nil,
		)
		return true
	}
	var targetObjectId string
	// I *think* the spec says that this must be an object. Not sure though
	switch target := rawTarget.(type) {
	case string:
		targetObjectId = target
	case map[string]any:
		objType, ok := target["type"].(string)
		// TODO: Ensure accept is only used for follows
		if !ok {
			log.Debug().Msg("Rejecting accept, missing embedded object data")
			_ = webutils.ProblemDetails(
				w,
				http.StatusBadRequest,
				"/errors/bad-request-data",
				"Bad activity data",
				other.IntoPointer(`Target object must contain a type field`),
				nil,
			)
			return true
		}
		if objType != "Follow" {
			log.Debug().Msg("Rejecting accept, bad object type")
			_ = webutils.ProblemDetails(
				w,
				http.StatusBadRequest,
				"/errors/bad-request-data",
				"Bad activity data",
				other.IntoPointer(`Target object type must be a string with value "Follow"`),
				nil,
			)
			return true
		}
		targetObjectId, ok = target["id"].(string)
		if !ok {
			log.Debug().Msg("Rejecting accept, missing object id")
			_ = webutils.ProblemDetails(
				w,
				http.StatusBadRequest,
				"/errors/bad-request-data",
				"Bad activity data",
				other.IntoPointer(`Missing id in undone object`),
				nil,
			)
			return true
		}
	default:
		log.Debug().Msg("Rejecting accept, missing object field")
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer(`Request data needs to contain a field "object" of type string or object`),
			nil,
		)
		return true
	}
	var internalId string
	if strings.HasPrefix(targetObjectId, "http") {
		internalIdMatch := objectIdRegex.FindStringSubmatch(targetObjectId)
		if len(internalIdMatch) != 2 {
			log.Debug().Msg("Rejecting accept, objectIdRegex didn't match")
			_ = webutils.ProblemDetails(
				w,
				http.StatusBadRequest,
				"/errors/bad-request-data",
				"Bad activity data",
				other.IntoPointer(`Request data target object is not internal id`),
				nil,
			)
			return true
		}
		internalId = internalIdMatch[1]
	} else {
		internalId = targetObjectId
	}
	followActivity, err := dbgen.Activity.Where(dbgen.Activity.Id.Eq(internalId)).First()
	switch err {
	case gorm.ErrRecordNotFound:
		log.Debug().Msg("Rejecting accept, target follow activity not found")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusNotFound)
		return true
	case nil:
	default:
		log.Error().
			Err(err).
			Str("target-id", internalId).
			Msg("Failed to get target follow activity from db")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	relationId := other.Must(strconv.ParseUint(followActivity.ObjectId, 10, 64))
	dbrel := dbgen.UserToUserRelation
	tx := dbgen.Q.Begin()
	_, err = tx.UserToUserRelation.Where(dbrel.ID.Eq(relationId)).
		UpdateColumn(dbrel.Relation, models.RelationFollow)
	switch err {
	case gorm.ErrRecordNotFound:
		// No need to rollback, nothing was done
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusNotFound)
		return true
	case nil:
	default:
		_ = tx.Rollback()
		log.Error().
			Err(err).
			Str("target-id", internalId).
			Msg("Failed to update follow status to confirmed follow")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	activity := models.Activity{
		Id:         object["id"].(string),
		Type:       string(models.ActivityAccept),
		ObjectType: uint32(models.ActivitystreamsActivityTargetActivity),
		ObjectId:   followActivity.Id,
	}
	err = tx.Activity.Create(&activity)
	if err != nil {
		err = tx.Rollback()
		log.Error().
			Err(err).
			Str("target-id", internalId).
			Msg("Failed to store accept activity in db")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	err = tx.Commit()
	if err != nil {
		log.Error().
			Err(err).
			Str("target-id", internalId).
			Msg("Failed to commit accept transaction")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	return true
}

// WARN: Untested as can't send follow activities yet
func handleReject(w http.ResponseWriter, r *http.Request, object map[string]any) bool {
	log := hlog.FromRequest(r)
	rawTarget, ok := object["object"]
	if !ok {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer(`Request data needs to contain a field "object"`),
			nil,
		)
		return true
	}
	// FIXME: Also handle other undo cases, such as follows
	var targetObjectId string
	// I *think* the spec says that this must be an object. Not sure though
	switch target := rawTarget.(type) {
	case string:
		targetObjectId = target
	case map[string]any:
		objType, ok := target["type"].(string)
		// TODO: Ensure accept is only used for follows
		if !ok || objType != "Follow" {
			_ = webutils.ProblemDetails(
				w,
				http.StatusBadRequest,
				"/errors/bad-request-data",
				"Bad activity data",
				other.IntoPointer(`Target object type must be a string with value "Follow"`),
				nil,
			)
			return true
		}
		targetObjectId, ok = target["id"].(string)
		if !ok {
			_ = webutils.ProblemDetails(
				w,
				http.StatusBadRequest,
				"/errors/bad-request-data",
				"Bad activity data",
				other.IntoPointer(`Missing id in undone object`),
				nil,
			)
			return true
		}
	default:
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer(`Request data needs to contain a field "object" of type string or object`),
			nil,
		)
		return true
	}
	internalIdMatch := objectIdRegex.FindStringSubmatch(targetObjectId)
	if len(internalIdMatch) != 2 {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"Bad activity data",
			other.IntoPointer(`Request data target object is not internal id`),
			nil,
		)
		return true
	}
	internalId := internalIdMatch[1]
	followActivity, err := dbgen.Activity.Where(dbgen.Activity.Id.Eq(internalId)).First()
	switch err {
	case gorm.ErrRecordNotFound:
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusNotFound)
		return true
	case nil:
	default:
		log.Error().
			Err(err).
			Str("target-id", internalId).
			Msg("Failed to get target follow activity from db")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	relationId := other.Must(strconv.ParseUint(followActivity.ObjectId, 10, 64))
	dbrel := dbgen.UserToUserRelation
	tx := dbgen.Q.Begin()
	_, err = tx.UserToUserRelation.Where(dbrel.ID.Eq(relationId)).Delete()
	switch err {
	case gorm.ErrRecordNotFound:
		// No need to rollback, nothing was done
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusNotFound)
		return true
	case nil:
	default:
		_ = tx.Rollback()
		log.Error().
			Err(err).
			Str("target-id", internalId).
			Msg("Failed to delete follow status")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	_, err = tx.Activity.Where(
		dbgen.Activity.ObjectId.Eq(followActivity.Id),
		dbgen.Activity.Type.Eq("Accept"),
	).
		Delete()
	if err != nil {
		_ = tx.Rollback()
		log.Error().Err(err).Msg("Failed to delete accept for later rejected follow")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	activity := models.Activity{
		Id:         object["id"].(string),
		Type:       string(models.ActivityAccept),
		ObjectType: uint32(models.ActivitystreamsActivityTargetActivity),
		ObjectId:   followActivity.Id,
	}
	err = tx.Activity.Create(&activity)
	if err != nil {
		err = tx.Rollback()
		log.Error().
			Err(err).
			Str("target-id", internalId).
			Msg("Failed to store accept activity in db")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	err = tx.Commit()
	if err != nil {
		log.Error().
			Err(err).
			Str("target-id", internalId).
			Msg("Failed to commit accept transaction")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	return true
}

func handleCreate(w http.ResponseWriter, r *http.Request, object map[string]any) bool {
	log := hlog.FromRequest(r)
	activity := translators.ActivityCreate{}
	err := mapstructure.Decode(object, &activity)
	targetUserId := r.PathValue("id")
	if err != nil {
		log.Error().
			Err(err).
			Any("raw", object).
			Msg("Failed to marshal create activity to proper type")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return true
	}
	actingUser, err := activitypub.ImportRemoteAccountByAPUrl(activity.Actor)
	if err != nil {
		log.Error().
			Err(err).
			Str("actor", activity.Actor).
			Msg("Failed to import remote actor for note")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return true
	}
	switch val := activity.Object.(type) {
	case string:
		actor, err := dbgen.User.GetById(targetUserId)
		if err != nil {
			log.Error().Err(err).Msg("Failed to get local actor for importing targeted note")
			_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
			return false
		}
		_, err = activitypub.ImportRemoteNote(val, actor)
		if err != nil {
			log.Error().Err(err).Str("note-url", val).Msg("Failed to import remote note that landed as id in the inbox")
			_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		}
		return false
	case map[string]any:
	default:
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"bad request data",
			other.IntoPointer("Bad object data for create activity. Must be a struct or string"),
			nil,
		)
		return true
	}
	obj := activity.Object.(map[string]any)
	// Dumb hack since published timestamp is still a string at this point
	tmpTime, err := time.Parse(time.RFC3339, obj["published"].(string))
	if err != nil {
		delete(obj, "published")
	} else {
		obj["published"] = tmpTime
	}
	objectNote := translators.ObjectNote{}
	err = mapstructure.Decode(obj, &objectNote)
	if err != nil {
		log.Error().
			Err(err).
			Any("raw", activity.Object).
			Msg("Failed to unmarshal create object into note")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return true
	}
	if objectNote.Type != "Note" {
		_ = webutils.ProblemDetails(
			w,
			http.StatusBadRequest,
			"/errors/bad-request-data",
			"bad request data",
			other.IntoPointer("Bad object data for create activity. object.type must be 'Note'"),
			nil,
		)
		return true
	}
	dbNote := models.Note{
		ID:         objectNote.Id,
		CreatedAt:  objectNote.Published,
		Creator:    actingUser,
		CreatorId:  actingUser.ID,
		Remote:     true,
		RawContent: objectNote.Content,
		OriginId:   actingUser.ServerId,
	}
	if objectNote.Summary != nil {
		dbNote.ContentWarning = sql.NullString{Valid: true, String: *objectNote.Summary}
	}
	if objectNote.InReplyTo != nil {
		replyUrl := *objectNote.InReplyTo
		if strings.HasPrefix(replyUrl, config.GlobalConfig.General.GetFullPublicUrl()) {
			// This is going to panic if the regex doesn't match. Who cares? Not me
			id := objectIdRegex.FindStringSubmatch(replyUrl)[1]
			dbNote.RepliesTo = sql.NullString{Valid: true, String: id}
		} else {
			dbNote.RepliesTo = sql.NullString{Valid: true, String: replyUrl}
		}
	}
	totalNoteTargets := slices.Concat(objectNote.To, objectNote.CC)
	// TODO: Find all feeds this note needs to be added to
	// Depends on note targets (cc & to tags)
	// Includes specific user IDs, nothing else -> DMs, only the personal feeds of those users
	// Includes https://www.w3.org/ns/activitystreams#Public (& creator follower list) ->
	//   Public feed (unlisted has same combo, but iirc doesn't get sent out by server)
	// Includes only creator follower list -> Follower only -> only feeds that explicitly follow creator

	u2u := dbgen.UserToUserRelation
	targetUserIds := []string{}
	if sliceutils.Contains(totalNoteTargets, "https://www.w3.org/ns/activitystreams#Public") {
		// Public post, add to global and following feeds
		dbNote.AccessLevel = models.NOTE_TARGET_PUBLIC
		followerIds, err := u2u.GetLocalFollowerIdsOfId(actingUser.ID)
		if err != nil {
			log.Error().
				Err(err).
				Str("follow-target", actingUser.ID).
				Msg("Failed to get ids for followers")
			return false
		}

		// globalFeed, err := dbgen.Feed.Where(dbgen.Feed.Name.Eq(models.GlobalFeedName)).First()
		// if err != nil {
		// 	log.Error().
		// 		Err(err).
		// 		Msg("Failed to get global feed")
		// 	return false
		// }
		targetUserIds = append(followerIds, storage.ServerActorId)
	} else {
		if sliceutils.ContainsFunc(totalNoteTargets, func(x string) bool {
			return strings.HasPrefix(x, actingUser.ID)
		}) {
			// Contains an url starting with the author's id, so assume that it's for followers
			dbNote.AccessLevel = models.NOTE_TARGET_FOLLOWERS
			followerIds, err := u2u.GetLocalFollowerIdsOfId(actingUser.ID)
			if err != nil {
				log.Error().
					Err(err).
					Str("follow-target", actingUser.ID).
					Msg("Failed to get ids for followers")
				return false
			}
			targetUserIds = followerIds
		} else {
			// Neither followers collection url nor public marker, private message
			dbNote.AccessLevel = models.NOTE_TARGET_DM
			urm := dbgen.UserRemoteLinks
			remoteLinks, err := urm.
				Join(dbgen.User, dbgen.User.ID.EqCol(urm.UserId)).
				Join(dbgen.RemoteServer, dbgen.RemoteServer.ID.EqCol(dbgen.User.ServerId)).
				Where(urm.ApLink.In(totalNoteTargets...), dbgen.RemoteServer.IsSelf.Is(true)).Find()
			if err != nil {
				log.Error().Err(err).Strs("target-ids", totalNoteTargets).Msg("Failed to get local targets for dm")
				return false
			}
			targetUserIds = sliceutils.Map(remoteLinks, func(t *models.UserRemoteLinks) string {
				return t.UserId
			})
		}
	}

	tx := dbgen.Q.Begin()
	err = tx.Note.Create(&dbNote)
	if err != nil {
		_ = tx.Rollback()
		log.Error().Err(err).Any("note", dbNote).Msg("Failed to create note in db")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	createActivity := models.Activity{
		Type:       string(models.ActivityCreate),
		Id:         object["id"].(string),
		ObjectId:   dbNote.ID,
		ObjectType: uint32(models.ActivitystreamsActivityTargetNote),
	}

	err = tx.Activity.Create(&createActivity)
	if err != nil {
		_ = tx.Rollback()
		log.Error().Err(err).Any("note", dbNote).Msg("Failed to create note create activity in db")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}

	feedRels := sliceutils.Map(targetUserIds, func(f string) *models.NoteToFeed {
		return &models.NoteToFeed{
			Reason:    string(models.FeedAppearanceReasonFollowUser),
			NoteId:    dbNote.ID,
			ForUserId: f,
		}
	})
	err = tx.NoteToFeed.Create(feedRels...)
	if err != nil {
		_ = tx.Rollback()
		log.Error().
			Err(err).
			Any("note", dbNote).
			Msg("Failed to create note to feed relations in db")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}

	err = tx.Commit()
	if err != nil {
		log.Error().Err(err).Any("note", dbNote).Msg("Failed to submit note creation in db")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return false
	}
	return true
}

func AcceptFollow(
	ctx context.Context,
	followRequestRelationId uint64,
	followerId, followedId string,
) error {
	u2u := dbgen.UserToUserRelation
	a := dbgen.Activity
	followRequestActivity, err := a.Where(a.ObjectId.Eq(fmt.Sprint(followRequestRelationId))).
		First()
	if err != nil {
		return other.Error("webaction", "failed to get follow activity", err)
	}
	follower, err := dbgen.User.GetById(followedId)
	if err != nil {
		return other.Error("webaction", "failed to get follower from db", err)
	}
	followed, err := dbgen.User.GetById(followedId)
	if err != nil {
		return err
	}
	tx := dbgen.Q.Begin()
	_, err = u2u.Where(u2u.ID.Eq(followRequestRelationId)).
		UpdateColumn(u2u.Relation, models.RelationFollow)
	if err != nil {
		_ = tx.Rollback()
		return err
	}
	acceptActivity := models.Activity{
		Id:         shared.NewId(),
		Type:       string(models.ActivityAccept),
		ObjectId:   followRequestActivity.Id,
		ObjectType: uint32(models.ActivitystreamsActivityTargetActivity),
	}
	err = tx.Activity.Create(&acceptActivity)
	if err != nil {
		_ = tx.Rollback()
		return err
	}
	err = tx.Commit()
	if err != nil {
		return err
	}
	webAccept, err := translators.AcceptFromStorage(ctx, acceptActivity.Id)
	if err != nil {
		return err
	}
	webAccept.Context = activitypub.BaseLdContext
	body, err := json.Marshal(webAccept)
	if err != nil {
		return err
	}
	res, _, err := webshared.RequestSigned(
		"POST",
		follower.RemoteInfo.InboxLink,
		body,
		followed,
	)
	if err != nil {
		return err
	}
	if res.StatusCode >= 400 {
		body, _ = io.ReadAll(res.Body)
		log.Error().
			Int("status-code", res.StatusCode).
			Bytes("body", body).
			Msg("Post of accept failed")
	}
	return nil
}
